package com.jiuqi.util.zip;

import java.io.UnsupportedEncodingException;
import java.util.Calendar;
import java.util.Date;
import java.util.zip.ZipException;
import java.util.zip.InflaterInputStream;
import java.util.zip.Inflater;
import java.io.InputStream;
import java.util.Enumeration;
import java.io.IOException;
import java.io.File;
import java.io.RandomAccessFile;
import java.util.Hashtable;

public class ZipFile
{
    private Hashtable entries;
    private Hashtable nameMap;
    private Hashtable dataOffsets;
    private String encoding;
    private RandomAccessFile archive;
    private static final int CFH_LEN = 42;
    private static final int MIN_EOCD_SIZE = 22;
    private static final int CFD_LOCATOR_OFFSET = 16;
    private static final long LFH_OFFSET_FOR_FILENAME_LENGTH = 26L;
    
    public ZipFile(final File f) throws IOException {
        this(f, null);
    }
    
    public ZipFile(final String name) throws IOException {
        this(new File(name), null);
    }
    
    public ZipFile(final String name, final String encoding) throws IOException {
        this(new File(name), encoding);
    }
    
    public ZipFile(final File f, final String encoding) throws IOException {
        this.entries = new Hashtable();
        this.nameMap = new Hashtable();
        this.dataOffsets = new Hashtable();
        this.encoding = null;
        this.encoding = encoding;
        this.archive = new RandomAccessFile(f, "r");
        this.populateFromCentralDirectory();
        this.resolveLocalFileHeaderData();
    }
    
    public String getEncoding() {
        return this.encoding;
    }
    
    public void close() throws IOException {
        this.archive.close();
    }
    
    public Enumeration getEntries() {
        return this.entries.keys();
    }
    
    public ZipEntry getEntry(final String name) {
        return this.nameMap.get(name);
    }
    
    public InputStream getInputStream(final ZipEntry ze) throws IOException, ZipException {
        final Long start = this.dataOffsets.get(ze);
        if (start == null) {
            return null;
        }
        final BoundedInputStream bis = new BoundedInputStream(start, ze.getCompressedSize());
        switch (ze.getMethod()) {
            case 0: {
                return bis;
            }
            case 8: {
                bis.addDummy();
                return new InflaterInputStream(bis, new Inflater(true));
            }
            default: {
                throw new ZipException("Found unsupported compression method " + ze.getMethod());
            }
        }
    }
    
    private void populateFromCentralDirectory() throws IOException {
        this.positionAtCentralDirectory();
        final byte[] cfh = new byte[42];
        final byte[] signatureBytes = new byte[4];
        this.archive.readFully(signatureBytes);
        for (ZipLong sig = new ZipLong(signatureBytes); sig.equals(ZipOutputStream.CFH_SIG); sig = new ZipLong(signatureBytes)) {
            this.archive.readFully(cfh);
            int off = 0;
            final ZipEntry ze = new ZipEntry();
            final ZipShort versionMadeBy = new ZipShort(cfh, off);
            off += 2;
            ze.setPlatform(versionMadeBy.getValue() >> 8 & 0xF);
            off += 4;
            ze.setMethod(new ZipShort(cfh, off).getValue());
            off += 2;
            ze.setTime(fromDosTime(new ZipLong(cfh, off)).getTime());
            off += 4;
            ze.setCrc(new ZipLong(cfh, off).getValue());
            off += 4;
            ze.setCompressedSize(new ZipLong(cfh, off).getValue());
            off += 4;
            ze.setSize(new ZipLong(cfh, off).getValue());
            off += 4;
            final int fileNameLen = new ZipShort(cfh, off).getValue();
            off += 2;
            final int extraLen = new ZipShort(cfh, off).getValue();
            off += 2;
            final int commentLen = new ZipShort(cfh, off).getValue();
            off += 2;
            off += 2;
            ze.setInternalAttributes(new ZipShort(cfh, off).getValue());
            off += 2;
            ze.setExternalAttributes(new ZipLong(cfh, off).getValue());
            off += 4;
            this.entries.put(ze, new Long(new ZipLong(cfh, off).getValue()));
            final byte[] fileName = new byte[fileNameLen];
            this.archive.readFully(fileName);
            ze.setName(this.getString(fileName));
            this.nameMap.put(ze.getName(), ze);
            this.archive.skipBytes(extraLen);
            final byte[] comment = new byte[commentLen];
            this.archive.readFully(comment);
            ze.setComment(this.getString(comment));
            this.archive.readFully(signatureBytes);
        }
    }
    
    private void positionAtCentralDirectory() throws IOException {
        long off = this.archive.length() - 22L;
        this.archive.seek(off);
        final byte[] sig = ZipOutputStream.EOCD_SIG.getBytes();
        int curr = this.archive.read();
        boolean found = false;
        while (curr != -1) {
            if (curr == sig[0]) {
                curr = this.archive.read();
                if (curr == sig[1]) {
                    curr = this.archive.read();
                    if (curr == sig[2]) {
                        curr = this.archive.read();
                        if (curr == sig[3]) {
                            found = true;
                            break;
                        }
                    }
                }
            }
            this.archive.seek(--off);
            curr = this.archive.read();
        }
        if (!found) {
            throw new ZipException("archive is not a ZIP archive");
        }
        this.archive.seek(off + 16L);
        final byte[] cfdOffset = new byte[4];
        this.archive.readFully(cfdOffset);
        this.archive.seek(new ZipLong(cfdOffset).getValue());
    }
    
    private void resolveLocalFileHeaderData() throws IOException {
        final Enumeration e = this.getEntries();
        while (e.hasMoreElements()) {
            final ZipEntry ze = e.nextElement();
            final long offset = this.entries.get(ze);
            this.archive.seek(offset + 26L);
            final byte[] b = new byte[2];
            this.archive.readFully(b);
            final int fileNameLen = new ZipShort(b).getValue();
            this.archive.readFully(b);
            final int extraFieldLen = new ZipShort(b).getValue();
            this.archive.skipBytes(fileNameLen);
            final byte[] localExtraData = new byte[extraFieldLen];
            this.archive.readFully(localExtraData);
            ze.setExtra(localExtraData);
            this.dataOffsets.put(ze, new Long(offset + 26L + 2L + 2L + fileNameLen + extraFieldLen));
        }
    }
    
    protected static Date fromDosTime(final ZipLong l) {
        final long dosTime = l.getValue();
        final Calendar cal = Calendar.getInstance();
        cal.set(1, (int)(dosTime >> 25 & 0x7FL) + 1980);
        cal.set(2, (int)(dosTime >> 21 & 0xFL) - 1);
        cal.set(5, (int)(dosTime >> 16) & 0x1F);
        cal.set(11, (int)(dosTime >> 11) & 0x1F);
        cal.set(12, (int)(dosTime >> 5) & 0x3F);
        cal.set(13, (int)(dosTime << 1) & 0x3E);
        return cal.getTime();
    }
    
    protected String getString(final byte[] bytes) throws ZipException {
        if (this.encoding == null) {
            return new String(bytes);
        }
        try {
            return new String(bytes, this.encoding);
        }
        catch (final UnsupportedEncodingException uee) {
            throw new ZipException(uee.getMessage());
        }
    }
    
    private class BoundedInputStream extends InputStream
    {
        private long remaining;
        private long loc;
        private boolean addDummyByte;
        
        BoundedInputStream(final long start, final long remaining) {
            this.addDummyByte = false;
            this.remaining = remaining;
            this.loc = start;
        }
        
        @Override
        public int read() throws IOException {
            if (this.remaining-- <= 0L) {
                if (this.addDummyByte) {
                    this.addDummyByte = false;
                    return 0;
                }
                return -1;
            }
            else {
                synchronized (ZipFile.this.archive) {
                    ZipFile.this.archive.seek(this.loc++);
                    final int read = ZipFile.this.archive.read();
                    monitorexit(ZipFile.this.archive);
                    return read;
                }
            }
        }
        
        @Override
        public int read(final byte[] b, final int off, int len) throws IOException {
            if (this.remaining <= 0L) {
                if (this.addDummyByte) {
                    this.addDummyByte = false;
                    b[off] = 0;
                    return 1;
                }
                return -1;
            }
            else {
                if (len <= 0) {
                    return 0;
                }
                if (len > this.remaining) {
                    len = (int)this.remaining;
                }
                int ret = -1;
                synchronized (ZipFile.this.archive) {
                    ZipFile.this.archive.seek(this.loc);
                    ret = ZipFile.this.archive.read(b, off, len);
                    monitorexit(ZipFile.this.archive);
                }
                if (ret > 0) {
                    this.loc += ret;
                    this.remaining -= ret;
                }
                return ret;
            }
        }
        
        void addDummy() {
            this.addDummyByte = true;
        }
    }
}
